try:
    from pyxmlhandler import _PyDictHandler
except ImportError:
    from .core import NativePyDictHandler as _PyDictHandler
from xml.parsers import expat


def parse(xml_content, attr_prefix: str = "@", cdata_key: str = "#text") -> dict:
    """
    Parse XML content into a python dictionary.

    Args:
        xml_content: The XML content to be parsed.
        attr_prefix: The prefix to use for attributes in the resulting dictionary.
        cdata_key: The key to use for character data in the resulting dictionary.

    Returns:
        A dictionary representation of the XML content.
    """
    handler = _PyDictHandler(attr_prefix=attr_prefix, cdata_key=cdata_key)
    parser = expat.ParserCreate()
    parser.CharacterDataHandler = handler.characters
    parser.StartElementHandler = handler.startElement
    parser.EndElementHandler = handler.endElement
    parser.Parse(xml_content, True)
    return handler.item


def parse_file(file_path, attr_prefix: str = "@", cdata_key: str = "#text") -> dict:
    """
    Parse an XML file into a python dictionary.

    Args:
        file_path: The path to the XML file to be parsed.
        attr_prefix: The prefix to use for attributes in the resulting dictionary.
        cdata_key: The key to use for character data in the resulting dictionary.

    Returns:
        A dictionary representation of the XML file content.
    """
    handler = _PyDictHandler(attr_prefix=attr_prefix, cdata_key=cdata_key)
    parser = expat.ParserCreate()
    parser.CharacterDataHandler = handler.characters
    parser.StartElementHandler = handler.startElement
    parser.EndElementHandler = handler.endElement
    with open(file_path, "rb") as f:
        parser.ParseFile(f)
    return handler.item


def iter_xml_documents(
    file_path, chunk_size: int = 64 * 1024, start_token: bytes = b"<?xml"
):
    buffer = b""
    with open(file_path, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                if buffer.strip():
                    yield buffer
                break
            buffer += chunk
            while True:
                start_index = buffer.find(start_token, 1)
                if start_index == -1:
                    break
                yield buffer[:start_index]
                buffer = buffer[start_index:]


def parse_xml_collections(
    file_path,
    attr_prefix: str = "@",
    cdata_key: str = "#text",
    start_token: bytes = b"<?xml",
):
    """
    Parse collections of xml documents based on a delimeter start_token

    Args:
        file_path: The path to the XML file to be parsed.
        attr_prefix: The prefix to use for attributes in the resulting dictionary.
        cdata_key: The key to use for character data in the resulting dictionary.
        start_token: The byte sequence that delimits the start of each XML document.

    Returns:
        A generator yielding dictionaries representing each XML document in the collection.
    """
    for xml_content in iter_xml_documents(file_path, start_token=start_token):
        yield parse(
            xml_content.decode("utf-8"), attr_prefix=attr_prefix, cdata_key=cdata_key
        )



import pytest
import json
from xmlpydict import parse, parse_file


def test_simple():
    assert parse("<p/>") == {"p": None}
    assert parse("<p></p>") == {"p": None}
    assert parse('<p width="10"></p>') == {"p": {"@width": "10"}}
    assert parse('<p width="10"></p>', attr_prefix="$") == {"p": {"$width": "10"}}
    assert parse("<p>Hello</p>") == {"p": "Hello"}

    assert parse('<p width="10">Hello World</p>') == {
        "p": {"@width": "10", "#text": "Hello World"}
    }
    assert parse('<p width="10" height="20"></p>') == {
        "p": {"@width": "10", "@height": "20"}
    }
    assert parse('<p width="10" height="20"/>') == {
        "p": {"@width": "10", "@height": "20"}
    }
    assert parse('<p width = "10"  height = "20"/>') == {
        "p": {"@width": "10", "@height": "20"}
    }
    assert parse("<p>Hey <b>bold</b>There</p>") == {
        "p": {"#text": "Hey There", "b": "bold"}
    }
    assert parse("<p>Hey <b>bold</b>There <b>bold</b>Buddy </p>") == {
        "p": {"#text": "Hey There Buddy", "b": ["bold", "bold"]}
    }

    assert parse("<p>Hey <b/>There Buddy</p>") == {
        "p": {"#text": "Hey There Buddy", "b": None}
    }

    assert parse("<p>Hey <b/>There Buddy <b/> </p>") == {
        "p": {"#text": "Hey There Buddy", "b": [None, None]}
    }

    assert (
        parse(
            """<svg width="200" height="100" xmlns="http://www.w3.org/2000/svg">
  <defs>
    <linearGradient id="gradient" x1="0%" y1="0%" x2="100%" y2="0%">
      <stop offset="0%" style="stop-color:rgb(255,255,0);stop-opacity:1" />
      <stop offset="100%" style="stop-color:rgb(255,0,0);stop-opacity:1" />
    </linearGradient>
  </defs>
  <rect width="200" height="100" fill="url(#gradient)" />
</svg>"""
        )
        == {
            "svg": {
                "@height": "100",
                "@width": "200",
                "@xmlns": "http://www.w3.org/2000/svg",
                "defs": {
                    "linearGradient": {
                        "@id": "gradient",
                        "@x1": "0%",
                        "@x2": "100%",
                        "@y1": "0%",
                        "@y2": "0%",
                        "stop": [
                            {
                                "@offset": "0%",
                                "@style": "stop-color:rgb(255,255,0);stop-opacity:1",
                            },
                            {
                                "@offset": "100%",
                                "@style": "stop-color:rgb(255,0,0);stop-opacity:1",
                            },
                        ],
                    }
                },
                "rect": {"@fill": "url(#gradient)", "@height": "100", "@width": "200"},
            }
        }
    )


def test_cdata():
    assert parse("<content><![CDATA[<p>This is a paragraph</p>]]></content>") == {
        "content": "<p>This is a paragraph</p>"
    }
    assert parse(
        "<special_chars><![CDATA[$ ^ * % & <> () + - + ` ~]]></special_chars>"
    ) == {"special_chars": "$ ^ * % & <> () + - + ` ~"}


def test_nested():
    assert parse("<book><p/></book> ") == {"book": {"p": None}}
    assert parse("<book><p></p></book>") == {"book": {"p": None}}


def test_list():
    xml_items = """<items>
  <item id="1"></item>
  <item id="2"></item>
  <item id="3"></item>
</items>"""
    assert parse(xml_items) == {
        "items": {"item": [{"@id": "1"}, {"@id": "2"}, {"@id": "3"}]}
    }


def test_comment():
    assert parse("<p/><!-- simple comment -->") == {"p": None}
    comment = """<world>
  <!-- $comment+++@python -->
  <lake>Content</lake>
</world>"""
    assert parse(comment) == {"world": {"lake": "Content"}}
    multiple_comments = """<book>
    <!-- Comment 0 -->
    <!-- Comment 1 -->
    <lines>510</lines>
    <!-- Comment 2 -->
    <!-- -->
</book>"""
    assert parse(multiple_comments) == {"book": {"lines": "510"}}


def test_files():
    svg_data = parse(
        """<svg xmlns="http://www.w3.org/2000/svg" width="400" height="400">
  <rect x="50" y="50" width="100" height="50" fill="blue" />
  <circle cx="200" cy="100" r="50" fill="red" />
  <ellipse cx="350" cy="75" rx="50" ry="25" fill="green" />
  <line x1="50" y1="200" x2="150" y2="300" stroke="orange" />
  <polyline points="200,200 250,250 300,200 350,250" fill="none" stroke="purple" />
  <polygon points="350,200 400,250 400,150" fill="yellow" />
  <path d="M50,350 L100,350 Q125,375 150,350 T200,350" fill="none" stroke="black"/>

  <rect x="10" y="10" height="100" width="100"
        style="stroke:#ff0000; fill: #0000ff"/>
        <path d="M50,350 L100,350 Q125,375 150,350 T200,350" fill="none" stroke="black"/><polygon points="350,200 400,250 400,150" fill="yellow" />


  <circle cx="200" cy="100" r="50" fill="red"></circle>

  <polygon points="350,200 400,250 400,150" fill="yellow" />
</svg>"""
    )

    json_data = json.loads(
        """{
    "svg": {
        "@xmlns": "http://www.w3.org/2000/svg",
        "@width": "400",
        "@height": "400",
        "rect": [
            {
                "@x": "50",
                "@y": "50",
                "@width": "100",
                "@height": "50",
                "@fill": "blue"
            },
            {
                "@x": "10",
                "@y": "10",
                "@height": "100",
                "@width": "100",
                "@style": "stroke:#ff0000; fill: #0000ff"
            }
        ],
        "circle": [
            {
                "@cx": "200",
                "@cy": "100",
                "@r": "50",
                "@fill": "red"
            },
            {
                "@cx": "200",
                "@cy": "100",
                "@r": "50",
                "@fill": "red"
            }
        ],
        "ellipse": {
            "@cx": "350",
            "@cy": "75",
            "@rx": "50",
            "@ry": "25",
            "@fill": "green"
        },
        "line": {
            "@x1": "50",
            "@y1": "200",
            "@x2": "150",
            "@y2": "300",
            "@stroke": "orange"
        },
        "polyline": {
            "@points": "200,200 250,250 300,200 350,250",
            "@fill": "none",
            "@stroke": "purple"
        },
        "polygon": [
            {
                "@points": "350,200 400,250 400,150",
                "@fill": "yellow"
            },
            {
                "@points": "350,200 400,250 400,150",
                "@fill": "yellow"
            },
            {
                "@points": "350,200 400,250 400,150",
                "@fill": "yellow"
            }
        ],
        "path": [
            {
                "@d": "M50,350 L100,350 Q125,375 150,350 T200,350",
                "@fill": "none",
                "@stroke": "black"
            },
            {
                "@d": "M50,350 L100,350 Q125,375 150,350 T200,350",
                "@fill": "none",
                "@stroke": "black"
            }
        ]
    }
}"""
    )

    assert svg_data == json_data

    svg_data = parse(
        """<svg xmlns="http://www.w3.org/2000/svg" width="200" height="200">
  <rect x="10" y="10" width="180" height="180" fill="lightgray" />
  
  <g transform="translate(100,100)">
    <circle cx="0" cy="0" r="80" fill="blue" />
    
    <g transform="translate(-40,-40)">
      <rect x="0" y="0" width="80" height="80" fill="red" />
      <text x="40" y="60" text-anchor="middle" fill="white">Nested</text>
    </g>
  </g>
</svg>"""
    )

    json_data = json.loads(
        """{
    "svg": {
        "@xmlns": "http://www.w3.org/2000/svg",
        "@width": "200",
        "@height": "200",
        "rect": {
            "@x": "10",
            "@y": "10",
            "@width": "180",
            "@height": "180",
            "@fill": "lightgray"
        },
        "g": {
            "@transform": "translate(100,100)",
            "circle": {
                "@cx": "0",
                "@cy": "0",
                "@r": "80",
                "@fill": "blue"
            },
            "g": {
                "@transform": "translate(-40,-40)",
                "rect": {
                    "@x": "0",
                    "@y": "0",
                    "@width": "80",
                    "@height": "80",
                    "@fill": "red"
                },
                "text": {
                    "@x": "40",
                    "@y": "60",
                    "@text-anchor": "middle",
                    "@fill": "white",
                    "#text": "Nested"
                }
            }
        }
    }
}"""
    )

    assert svg_data == json_data


def test_exception():
    xml_strings = [
        "<p height'10'/>",
        "<p height='10'width='5'/>",
        "<p width=5'/>",
        "<pwidth='5'/>",
        "<!---->",
        "<a></p>",
        "<></>",
        "</>",
        "<",
        ">",
        "<nested></p></nested>",
    ]
    for xml_str in xml_strings:
        with pytest.raises(Exception):
            parse(xml_str)


def test_prefix():
    assert parse("<p></p>", attr_prefix="$") == {"p": None}
    assert parse('<p width="10"></p>', attr_prefix="$") == {"p": {"$width": "10"}}
    assert parse('<p width="10" height="5"></p>', attr_prefix="$") == {
        "p": {"$width": "10", "$height": "5"}
    }
    assert parse('<p width="10" height="5"></p>', attr_prefix="$$$$$$$$$") == {
        "p": {"$$$$$$$$$width": "10", "$$$$$$$$$height": "5"}
    }
    assert parse('<p width="10" height="5"></p>', attr_prefix="") == {
        "p": {"width": "10", "height": "5"}
    }


def test_document():
    s = """<?xml version="1.0" encoding="UTF-8"?><repository>
  <project pypi="xmlpydict">
    <title>XML document parser</title>
    <author>Matthew Taylor</author>
  </project>
  <project pypi="blank">
    <title>Test project</title>
    <author>Matthew Taylor</author>
  </project>
</repository>"""
    assert parse(s) == {
        "repository": {
            "project": [
                {
                    "@pypi": "xmlpydict",
                    "title": "XML document parser",
                    "author": "Matthew Taylor",
                },
                {
                    "@pypi": "blank",
                    "title": "Test project",
                    "author": "Matthew Taylor",
                },
            ]
        }
    }


def test_parse_file(tmp_path):
    s = """<?xml version="1.0" encoding="UTF-8"?><repository>
  <project pypi="xmlpydict">
    <title>XML document parser</title>
    <author>Matthew Taylor</author>
  </project>
  <project pypi="blank">
    <title>Test project</title>
    <author>Matthew Taylor</author>
  </project>
</repository>"""
    with open(tmp_path / "test.xml", "w") as f:
        f.write(s)

    assert parse_file(tmp_path / "test.xml") == {
        "repository": {
            "project": [
                {
                    "@pypi": "xmlpydict",
                    "title": "XML document parser",
                    "author": "Matthew Taylor",
                },
                {
                    "@pypi": "blank",
                    "title": "Test project",
                    "author": "Matthew Taylor",
                },
            ]
        }
    }



# Python version of PyDictHandler


class NativePyDictHandler:
    """
    Handler class for parsing XML content into a Python dictionary using the expat parser.
    """
    
    def __init__(
        self,
        attr_prefix: str = "@",
        cdata_key: str = "#text",
    ):
        self.attr_prefix = attr_prefix
        self.cdata_key = cdata_key

        self.item = None
        # Use list instead of string as in cpp version for efficiency
        self.data = []

        self.item_stack = []
        self.data_stack = []

    def characters(self, data):
        self.data.append(data)

    def startElement(self, name, attrs):
        self.item_stack.append(self.item)
        self.data_stack.append(self.data)
        new_dict = {self.attr_prefix + key: value for key, value in attrs.items()}
        self.item = new_dict or None
        self.data = []

    @staticmethod
    def update_children(target, key, value):
        if target is None:
            target = {}

        if not key in target:
            target[key] = value
        else:
            existing = target[key]
            if isinstance(existing, list):
                existing.append(value)
            else:
                target[key] = [existing, value]
        return target

    def endElement(self, name):
        if self.data_stack:
            py_data = "".join(self.data).strip() or None
            temp_item = self.item

            self.item = self.item_stack.pop()
            self.data = self.data_stack.pop()

            if temp_item is not None:
                if py_data:
                    temp_item[self.cdata_key] = py_data
                self.item = self.update_children(self.item, name, temp_item)
            else:
                self.item = self.update_children(self.item, name, py_data)



from setuptools import setup, Extension
from distutils.command.build_ext import build_ext as build_ext_orig


class build_ext(build_ext_orig):
    def build_extension(self, ext):
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        return super().get_ext_filename(ext_name)


setup(
    include_package_data=True,
    ext_modules=[
        Extension("pyxmlhandler", ["src/xmlparse.cpp"]),
    ],
    cmdclass={"build_ext": build_ext},
    packages=["xmlpydict"],
)
